/*
    libtimeslide -- add a constant to the system time
    Copyright (C) 2003  Egmont Koblinger <egmont@uhulinux.hu>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#define _GNU_SOURCE
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <time.h>
#include <utime.h>
#include <errno.h>

static time_t (*next_time)(time_t *);
static int (*next_ftime)(struct timeb *);
static int (*next_stime)(const time_t *);
static int (*next_gettimeofday)(struct timeval*, struct timezone *);
static int (*next_settimeofday)(const struct timeval *tv, const struct timezone *tz);
static int (*next_clock_gettime)(clockid_t, struct timespec *);
static int (*next_clock_settime)(clockid_t, const struct timespec *);
static int (*next___xstat)(int, const char *, struct stat *);
static int (*next___lxstat)(int, const char *, struct stat *);
static int (*next___fxstat)(int, int, struct stat *);
static int (*next___xstat64)(int, const char *, struct stat64 *);
static int (*next___lxstat64)(int, const char *, struct stat64 *);
static int (*next___fxstat64)(int, int, struct stat64 *);
static int (*next_utime)(const char *, const struct utimbuf *);
static int (*next_utimes)(const char *, const struct timeval *);
static int (*next_lutimes)(const char *, const struct timeval *);
static int (*next_futimes)(int, const struct timeval *);

static long slide = 0;
static int trivial = 0;

void timeslide_init (void) __attribute((constructor));
void timeslide_init (void)
{
	char *s;
	char dummy;

	next_time          = dlsym(RTLD_NEXT, "time");
	next_ftime         = dlsym(RTLD_NEXT, "ftime");
	next_stime         = dlsym(RTLD_NEXT, "stime");
	next_gettimeofday  = dlsym(RTLD_NEXT, "gettimeofday");
	next_settimeofday  = dlsym(RTLD_NEXT, "settimeofday");
	next_clock_gettime = dlsym(RTLD_NEXT, "clock_gettime");
	next_clock_settime = dlsym(RTLD_NEXT, "clock_settime");
	next___xstat       = dlsym(RTLD_NEXT, "__xstat");
	next___lxstat      = dlsym(RTLD_NEXT, "__lxstat");
	next___fxstat      = dlsym(RTLD_NEXT, "__fxstat");
	next___xstat64     = dlsym(RTLD_NEXT, "__xstat64");
	next___lxstat64    = dlsym(RTLD_NEXT, "__lxstat64");
	next___fxstat64    = dlsym(RTLD_NEXT, "__fxstat64");
	next_utime         = dlsym(RTLD_NEXT, "utime");
	next_utimes        = dlsym(RTLD_NEXT, "utimes");
	next_lutimes       = dlsym(RTLD_NEXT, "lutimes");
	next_futimes       = dlsym(RTLD_NEXT, "futimes");

	s = getenv("TIMESLIDE");
	if (s != NULL) {
		if (sscanf(s, "%ld%c", &slide, &dummy) != 1) {
			slide = 0;
		}
	}

	if (getenv("TIMESLIDE_TRIVIAL") == NULL) trivial = 1;
}

time_t time (time_t *t)
{
	time_t t2;
	t2 = next_time(t) + slide;
	if (t != NULL) *t = t2;
	return t2;
}

int ftime (struct timeb *tp)
{
	int retval;
	retval = next_ftime(tp);
	if (retval >= 0) tp->time += slide;
	return retval;
}

int stime (const time_t *t)
{
	time_t t2;
	t2 = *t - slide;
	return next_stime(&t2);
}

int gettimeofday (struct timeval *tv, struct timezone *tz)
{
	int retval;
	retval = next_gettimeofday(tv, tz);
	if (retval >= 0) tv->tv_sec += slide;
	return retval;
}

int settimeofday (const struct timeval *tv, const struct timezone *tz)
{
	struct timeval tv2;
	tv2.tv_sec = tv->tv_sec - slide;
	tv2.tv_usec = tv->tv_usec;
	return next_settimeofday(&tv2, tz);
}

int clock_gettime (clockid_t clk_id, struct timespec *tp)
{
	int retval;
	retval = next_clock_gettime(clk_id, tp);
	if (retval >= 0 && clk_id == CLOCK_REALTIME) tp->tv_sec += slide;
	return retval;
}

int clock_settime (clockid_t clk_id, const struct timespec *tp)
{
	struct timespec tp2;
	tp2.tv_sec = tp->tv_sec - (clk_id == CLOCK_REALTIME ? slide : 0);
	tp2.tv_nsec = tp->tv_nsec;
	return next_clock_settime(clk_id, &tp2);
}

int __xstat (int ver, const char *file, struct stat *st)
{
	int retval;
	retval = next___xstat(ver, file, st);
	if (trivial || retval < 0) return retval;
	st->st_atime += slide;
	st->st_ctime += slide;
	st->st_mtime += slide;
	return 0;
}

int __lxstat (int ver, const char *file, struct stat *st)
{
	int retval;
	retval = next___lxstat(ver, file, st);
	if (trivial || retval < 0) return retval;
	st->st_atime += slide;
	st->st_ctime += slide;
	st->st_mtime += slide;
	return 0;
}

int __fxstat (int ver, int fd, struct stat *st)
{
	int retval;
	retval = next___fxstat(ver, fd, st);
	if (trivial || retval < 0) return retval;
	st->st_atime += slide;
	st->st_ctime += slide;
	st->st_mtime += slide;
	return 0;
}

int __xstat64 (int ver, const char *file, struct stat64 *st)
{
	int retval;
	retval = next___xstat64(ver, file, st);
	if (trivial || retval < 0) return retval;
	st->st_atime += slide;
	st->st_ctime += slide;
	st->st_mtime += slide;
	return 0;
}

int __lxstat64 (int ver, const char *file, struct stat64 *st)
{
	int retval;
	retval = next___lxstat64(ver, file, st);
	if (trivial || retval < 0) return retval;
	st->st_atime += slide;
	st->st_ctime += slide;
	st->st_mtime += slide;
	return 0;
}

int __fxstat64 (int ver, int fd, struct stat64 *st)
{
	int retval;
	retval = next___fxstat64(ver, fd, st);
	if (trivial || retval < 0) return retval;
	st->st_atime += slide;
	st->st_ctime += slide;
	st->st_mtime += slide;
	return 0;
}

int utime (const char *file, const struct utimbuf *buf)
{
	struct utimbuf buf2;
	if (trivial || buf == NULL) return next_utime(file, buf);
	buf2.actime = buf->actime - slide;
	buf2.modtime = buf->modtime - slide;
	return next_utime(file, &buf2);
}

int utimes (const char *file, const struct timeval *tvp)
{
	struct timeval tv2[2];
	if (trivial) return next_utimes(file, tvp);
	tv2[0].tv_sec = tvp[0].tv_sec - slide;
	tv2[0].tv_usec = tvp[0].tv_usec;
	tv2[1].tv_sec = tvp[1].tv_sec - slide;
	tv2[1].tv_usec = tvp[1].tv_usec;
	return next_utimes(file, tv2);
}

int lutimes (const char *file, const struct timeval *tvp)
{
	struct timeval tv2[2];
	if (trivial) return next_lutimes(file, tvp);
	tv2[0].tv_sec = tvp[0].tv_sec - slide;
	tv2[0].tv_usec = tvp[0].tv_usec;
	tv2[1].tv_sec = tvp[1].tv_sec - slide;
	tv2[1].tv_usec = tvp[1].tv_usec;
	return next_lutimes(file, tv2);
}

int futimes (int fd, const struct timeval *tvp)
{
	struct timeval tv2[2];
	if (trivial) return next_futimes(fd, tvp);
	tv2[0].tv_sec = tvp[0].tv_sec - slide;
	tv2[0].tv_usec = tvp[0].tv_usec;
	tv2[1].tv_sec = tvp[1].tv_sec - slide;
	tv2[1].tv_usec = tvp[1].tv_usec;
	return next_futimes(fd, tv2);
}

